-module(driver).
-export([driver/6, parallel_driver/7, eval_fitness_individual/4]).

%
% driver : main routines for the process
%
% matt@cs.uoregon.edu
%

%
% populate a dictionary from a list of key/value pairs
%
populate_dict([],D) -> D;
populate_dict([{X,V}|T], D) -> populate_dict(T, dict:store(X,V,D)).

%
% evaluate the fitness of an individual given the individual, the
% maximum fitness, the list of input key/value pairs, and the
% set of outputs for each test
%
eval_fitness_individual(_, _, [], _) -> 0.0;
eval_fitness_individual(Individual, M, [H|T], [HV|TV]) ->
  DataDict = populate_dict(H,dict:new()),
  Expressed = express:express(Individual),
  try gaeval:eval(hd(Expressed),DataDict) of
    EvalResult -> ((M-abs(HV-EvalResult)) + 
      (eval_fitness_individual(Individual,M,T,TV)))
  catch
    _:_ -> -1000000.0 + (eval_fitness_individual(Individual,M,T,TV))
  end.

%
% make a new population from an old one with the selections provided
%
make_new_pop(_,[],_) -> [];
make_new_pop(SortedPop,Selections,Hlength) ->
  I1 = lists:nth(hd(Selections),SortedPop),
  I2 = lists:nth(hd(tl(Selections)),SortedPop),
  [gaops:breed(I1,I2,Hlength,0.01,0.5)]++make_new_pop(SortedPop,tl(tl(Selections)),Hlength).

%
% make sublists for workers
% 
sublists(1,_,L) -> [L];
sublists(N,Size,L) ->
  {A,B} = lists:split(Size,L),
  [A]++(sublists(N-1,Size,B)).
 
%
% gather evaulation results from workers
gather_results(0,R,P) -> {R,P};
gather_results(N,R,P) ->
  receive
    {Pid,{Results,SubPop}} -> 
      gather_results(N-1,R++Results,P++SubPop)
  end.
  
%
% spawn worker threads to do the fitness evaluation
%
parallel_fitness_eval(Pids,Population,M,DataPoints,DataValues) ->
  NumPids = length(Pids),
  NumIndividuals = length(Population),
  SubPops = sublists(NumPids,round(NumIndividuals/NumPids),Population),
  lists:map(fun({I,Pid}) -> 
            Pid ! {self(), {work,{I,M,DataPoints,DataValues}}} end, 
            lists:zip(SubPops,Pids)),
  gather_results(NumPids,[],[]).

parallel_iterator(0,Population,_,_,_,_,_) -> hd(Population);
parallel_iterator(Niters, OrigPop, M, Hlength, DataPoints, 
                  DataValues, Pids) ->
  {Results,Population} = parallel_fitness_eval(Pids,OrigPop,M,DataPoints,
                                               DataValues),
  SortedPairs = 
      lists:sort(fun({R1,_},{R2,_}) -> 
                   if 
                     (R1 > R2) -> true;
                     true -> false
                   end end,
                 lists:zip(Results,Population)),
  {Fitnesses,Individuals} = lists:unzip(SortedPairs),
  io:format("BEST FITNESS:~p ~p~n",[Niters, hd(Fitnesses)]),
  N = length(Individuals),
  W1 = gaselect:roulette_wheel(N,
         gaselect:generate_weights(N,1.0,0.9,0.01)),
  W2 = gaselect:roulette_wheel(N,
         gaselect:generate_weights(N,1.0,0.9,0.01)),
  W = W1++tl(tl(W2)),
  NewPop = [hd(Individuals)]++(make_new_pop(Individuals,W,Hlength)),
  parallel_iterator(Niters-1,NewPop,M,Hlength,DataPoints,DataValues,Pids).

  
iterator(0,Population,_,_,_,_) -> hd(Population);
iterator(Niters, Population, M, Hlength, DataPoints, DataValues) ->
  Results = lists:map(fun(I) ->
    eval_fitness_individual(I,M,DataPoints,DataValues) end, Population),
  SortedPairs = 
      lists:sort(fun({R1,_},{R2,_}) -> 
                   if 
                     (R1 > R2) -> true;
                     true -> false
                   end end,
                 lists:zip(Results,Population)),
  {Fitnesses,Individuals} = lists:unzip(SortedPairs),
  io:format("BEST FITNESS:~p ~p~n",[Niters, hd(Fitnesses)]),
  N = length(Individuals),
  W1 = gaselect:roulette_wheel(N,
      gaselect:generate_weights(N,1.0,0.9,0.01)),
  W2 = gaselect:roulette_wheel(N,
          gaselect:generate_weights(N,1.0,0.9,0.01)),
  W = W1++tl(tl(W2)),
  NewPop = [hd(Individuals)]++(make_new_pop(Individuals,W,Hlength)),
  iterator(Niters-1,NewPop,M,Hlength,DataPoints,DataValues).

make_pids(0) -> [];
make_pids(N) -> [(spawn(fun worker:loop/0))]++(make_pids(N-1)).

parallel_driver(Nindividuals, Niter, Hlength, M, DataPoints, DataValues, NP) ->
  Pop = gagenome:random_population(Nindividuals,Hlength),
  Pids = make_pids(NP),
  parallel_iterator(Niter,Pop,M,Hlength,DataPoints,DataValues,Pids).

driver(Nindividuals, Niter, Hlength, M, DataPoints, DataValues) ->
  Pop = gagenome:random_population(Nindividuals, Hlength),
  iterator(Niter,Pop,M,Hlength,DataPoints,DataValues).